@@ -5,17 +5,18 @@ from decimal import Decimal
 from itertools import groupby
 from functools import partial
 
+from trytond.i18n import gettext
 from trytond.model import ModelView, ModelSQL, fields
+from trytond.model.exceptions import AccessError
 from trytond.wizard import Wizard, StateView, StateTransition, Button
 from trytond.pyson import If, In, Eval, Bool
+from trytond.tools import sortable_values, firstline
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
-__all__ = ['PurchaseRequest',
-    'CreatePurchaseAskParty', 'CreatePurchase',
-    'HandlePurchaseCancellationException',
-    'HandlePurchaseCancellationExceptionStart',
-    ]
+from trytond.modules.company.model import (
+    employee_field, set_employee, reset_employee)
+from trytond.modules.product import round_price
 
 STATES = {
     'readonly': Eval('state') != 'draft',
@@ -27,27 +28,48 @@ class PurchaseRequest(ModelSQL, ModelView):
     'Purchase Request'
     __name__ = 'purchase.request'
 
-    product = fields.Many2One('product.product', 'Product',
-        select=True, readonly=True, domain=[('purchasable', '=', True)])
+    product = fields.Many2One(
+        'product.product', "Product", select=True, readonly=True,
+        domain=[('purchasable', '=', True)],
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=['company'])
     description = fields.Text('Description', readonly=True,
         states=STATES, depends=DEPENDS)
-    party = fields.Many2One('party.party', 'Party', select=True, states=STATES,
-        depends=DEPENDS)
+    summary = fields.Function(fields.Char('Summary'), 'on_change_with_summary')
+    party = fields.Many2One(
+        'party.party', "Party", select=True, states=STATES,
+        context={
+            'company': Eval('company', -1),
+            },
+        depends=DEPENDS + ['company'])
     quantity = fields.Float('Quantity', required=True, states=STATES,
         digits=(16, Eval('uom_digits', 2)), depends=DEPENDS + ['uom_digits'])
     uom = fields.Many2One('product.uom', 'UOM', select=True,
+        ondelete='RESTRICT',
+        domain=[
+            If(Bool(Eval('product_uom_category')),
+                ('category', '=', Eval('product_uom_category')),
+                ('category', '!=', -1)),
+            ],
         states={
             'required': Bool(Eval('product')),
             'readonly': STATES['readonly'],
             },
-        depends=['product'] + DEPENDS)
+        depends=['product', 'product_uom_category'] + DEPENDS)
     uom_digits = fields.Function(fields.Integer('UOM Digits'),
         'on_change_with_uom_digits')
+    product_uom_category = fields.Function(
+        fields.Many2One('product.uom.category', "Product Uom Category"),
+        'on_change_with_product_uom_category')
     computed_quantity = fields.Float('Computed Quantity', readonly=True)
     computed_uom = fields.Many2One('product.uom', 'Computed UOM',
         readonly=True)
     purchase_date = fields.Date('Best Purchase Date', readonly=True)
     supply_date = fields.Date('Expected Supply Date', readonly=True)
+    default_uom = fields.Function(fields.Many2One(
+            'product.uom', "Default UOM"), 'on_change_with_default_uom')
     default_uom_digits = fields.Function(fields.Integer('Default UOM Digits'),
         'on_change_with_default_uom_digits')
     stock_level = fields.Float('Stock at Supply Date', readonly=True,
@@ -73,11 +95,14 @@ class PurchaseRequest(ModelSQL, ModelView):
             ])
     origin = fields.Reference('Origin', selection='get_origin', readonly=True)
     exception_ignored = fields.Boolean('Ignored Exception')
+
+    purchased_by = employee_field(
+        "Purchased By", states=['purchased', 'done', 'cancelled', 'exception'])
     state = fields.Selection([
             ('purchased', "Purchased"),
             ('done', "Done"),
             ('draft', "Draft"),
-            ('cancel', "Cancel"),
+            ('cancelled', "Cancelled"),
             ('exception', "Exception"),
             ], "State", required=True, readonly=True, select=True)
 
@@ -85,12 +110,6 @@ class PurchaseRequest(ModelSQL, ModelView):
     def __setup__(cls):
         super(PurchaseRequest, cls).__setup__()
         cls._order[0] = ('id', 'DESC')
-        cls._error_messages.update({
-                'create_request': ('Purchase requests are only created '
-                    'by the system.'),
-                'delete_purchase_line': ('You can not delete purchased '
-                    'request.'),
-                })
         cls._buttons.update({
                 'handle_purchase_cancellation_exception': {
                     'invisible': Eval('state') != 'exception',
@@ -158,13 +177,28 @@ class PurchaseRequest(ModelSQL, ModelView):
         # Migration from 4.4: remove required on origin
         tablehandler.not_null_action('origin', action='remove')
 
+        # Migration from 5.6: rename state cancel to cancelled
+        cursor.execute(*request.update(
+                [request.state], ['cancelled'],
+                where=request.state == 'cancel'))
+
     def get_rec_name(self, name):
-        product_name = (self.product.name if self.product else
-            self.description.splitlines()[0])
-        if self.warehouse:
-            return "%s@%s" % (product_name, self.warehouse.name)
+        pool = Pool()
+        Lang = pool.get('ir.lang')
+        if self.product:
+            lang = Lang.get()
+            rec_name = (lang.format(
+                    '%.*f', (self.uom.digits, self.quantity))
+                + '%s %s' % (self.uom.symbol, self.product.name))
+        elif self.description:
+            rec_name = self.description.splitlines()[0]
         else:
-            return product_name
+            rec_name = str(self.id)
+
+        if self.warehouse:
+            return "%s @% s" % (rec_name, self.warehouse.name)
+        else:
+            return rec_name
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -203,11 +237,11 @@ class PurchaseRequest(ModelSQL, ModelView):
 
     def get_state(self):
         if self.purchase_line:
-            if (self.purchase_line.purchase.state == 'cancel'
+            if (self.purchase_line.purchase.state == 'cancelled'
                     and not self.exception_ignored):
                 return 'exception'
-            elif self.purchase_line.purchase.state == 'cancel':
-                return 'cancel'
+            elif self.purchase_line.purchase.state == 'cancelled':
+                return 'cancelled'
             elif self.purchase_line.purchase.state == 'done':
                 return 'done'
             else:
@@ -222,6 +256,16 @@ class PurchaseRequest(ModelSQL, ModelView):
                 request.state = state
         cls.save(requests)
 
+    @classmethod
+    @set_employee('purchased_by')
+    def set_purchased(cls, requests):
+        cls.update_state(requests)
+
+    @classmethod
+    @reset_employee('purchased_by')
+    def reset_purchased(cls, requests):
+        cls.update_state(requests)
+
     def get_warehouse_required(self, name):
         return self.product and self.product.type in ('goods', 'assets')
 
@@ -231,12 +275,26 @@ class PurchaseRequest(ModelSQL, ModelView):
             return self.uom.digits
         return 2
 
+    @fields.depends('product')
+    def on_change_with_product_uom_category(self, name=None):
+        if self.product:
+            return self.product.default_uom_category.id
+
+    @fields.depends('product')
+    def on_change_with_default_uom(self, name=None):
+        if self.product:
+            return self.product.default_uom.id
+
     @fields.depends('product')
     def on_change_with_default_uom_digits(self, name=None):
         if self.product:
             return self.product.default_uom.digits
         return 2
 
+    @fields.depends('description')
+    def on_change_with_summary(self, name=None):
+        return firstline(self.description or '')
+
     @classmethod
     def _get_origin(cls):
         'Return the set of Model names for origin Reference'
@@ -246,45 +304,73 @@ class PurchaseRequest(ModelSQL, ModelView):
     def get_origin(cls):
         pool = Pool()
         IrModel = pool.get('ir.model')
-        models = IrModel.search([
-                ('model', 'in', list(cls._get_origin())),
-                ])
-        return [(None, '')] + [(m.model, m.name) for m in models]
+        get_name = IrModel.get_name
+        models = cls._get_origin()
+        return [(None, '')] + [(m, get_name(m)) for m in models]
+
+    @classmethod
+    def view_attributes(cls):
+        return super().view_attributes() + [
+            ('/tree', 'visual', If(Eval('state') == 'cancelled', 'muted', '')),
+            ]
 
     @classmethod
     def create(cls, vlist):
         for vals in vlist:
             for field_name in ('quantity', 'company'):
                 if vals.get(field_name) is None:
-                    cls.raise_user_error('create_request')
+                    raise AccessError(
+                        gettext('purchase_request.msg_request_no_create'))
         return super(PurchaseRequest, cls).create(vlist)
 
     @classmethod
     def delete(cls, requests):
-        if any(r.purchase_line for r in requests):
-            cls.raise_user_error('delete_purchase_line')
+        for request in requests:
+            if request.purchase_line:
+                raise AccessError(
+                    gettext('purchase_request.msg_request_delete_purchased',
+                        request=request.rec_name))
         super(PurchaseRequest, cls).delete(requests)
 
     @classmethod
-    def find_best_supplier(cls, product, date):
+    def copy(cls, requests, default=None):
+        if default is None:
+            default = {}
+        else:
+            default = default.copy()
+        default.setdefault('purchased_by')
+        return super().copy(requests, default=default)
+
+    @classmethod
+    def find_best_product_supplier(cls, product, date, **pattern):
+        "Return the best product supplier to request product at date"
+        pool = Pool()
+        Date = pool.get('ir.date')
+        today = Date.today()
+        for product_supplier in product.product_suppliers_used(**pattern):
+            if date is None:
+                return product_supplier
+            supply_date = product_supplier.compute_supply_date(date=today)
+            timedelta = date - supply_date
+            if timedelta >= datetime.timedelta(0):
+                return product_supplier
+
+    @classmethod
+    def find_best_supplier(cls, product, date, **pattern):
         '''
         Return the best supplier and purchase_date for the product.
         '''
-        Date = Pool().get('ir.date')
+        pool = Pool()
+        Date = pool.get('ir.date')
 
-        supplier = None
-        today = Date.today()
-        for product_supplier in product.product_suppliers:
-            supply_date = product_supplier.compute_supply_date(date=today)
-            timedelta = date - supply_date
-            if not supplier and timedelta >= datetime.timedelta(0):
-                supplier = product_supplier.party
-                break
-
-        if supplier:
+        product_supplier = cls.find_best_product_supplier(
+            product, date, **pattern)
+        if product_supplier:
+            supplier = product_supplier.party
             purchase_date = product_supplier.compute_purchase_date(date)
         else:
-            purchase_date = today
+            supplier = None
+            purchase_date = Date.today()
         return supplier, purchase_date
 
     @classmethod
@@ -313,19 +399,8 @@ class CreatePurchase(Wizard):
             Button('Continue', 'start', 'tryton-forward', default=True),
             ])
 
-    @classmethod
-    def __setup__(cls):
-        super(CreatePurchase, cls).__setup__()
-        cls._error_messages.update({
-                'missing_price': 'Purchase price is missing for product "%s".',
-                'please_update': ('This price is necessary for creating '
-                    'purchases.'),
-                })
-
     def default_ask_party(self, fields):
-        Request = Pool().get('purchase.request')
-        requests = Request.browse(Transaction().context['active_ids'])
-        for request in requests:
+        for request in self.records:
             if request.purchase_line:
                 continue
             if not request.party:
@@ -336,8 +411,8 @@ class CreatePurchase(Wizard):
             'company': request.company.id,
             }
 
-    @staticmethod
-    def _group_purchase_key(requests, request):
+    @classmethod
+    def _group_purchase_key(cls, requests, request):
         '''
         The key to group lines by purchases
         A list of key-value as tuples of the purchase
@@ -369,7 +444,7 @@ class CreatePurchase(Wizard):
         Line = pool.get('purchase.line')
         Date = pool.get('ir.date')
 
-        requests = Request.browse(Transaction().context['active_ids'])
+        requests = self.records
 
         if (getattr(self.ask_party, 'party', None)
                 and getattr(self.ask_party, 'company', None)):
@@ -403,7 +478,7 @@ class CreatePurchase(Wizard):
         requests = [r for r in requests if not r.purchase_line]
 
         keyfunc = partial(self._group_purchase_key, requests)
-        requests = sorted(requests, key=keyfunc)
+        requests = sorted(requests, key=sortable_values(keyfunc))
 
         purchases = []
         lines = []
@@ -431,7 +506,7 @@ class CreatePurchase(Wizard):
                 lines.append(line)
         Purchase.save(purchases)
         Line.save(lines)
-        Request.update_state(requests)
+        Request.set_purchased(requests)
         return 'end'
 
     @classmethod
@@ -440,13 +515,16 @@ class CreatePurchase(Wizard):
         Line = pool.get('purchase.line')
 
         line = Line()
-        line.unit_price = Decimal(0).quantize(
-            Decimal(1) / 10 ** Line.unit_price.digits[1])
+        line.unit_price = round_price(Decimal(0))
         for f, v in key:
             setattr(line, f, v)
         line.quantity = sum(r.quantity for r in requests)
         line.purchase = purchase
         line.on_change_product()
+        # Set again in case on_change's changed them
+        for f, v in key:
+            setattr(line, f, v)
+        line.on_change_quantity()
         return line
 
 
@@ -465,23 +543,15 @@ class HandlePurchaseCancellationException(Wizard):
     cancel_request = StateTransition()
 
     def transition_reset(self):
-        pool = Pool()
-        PurchaseRequest = pool.get('purchase.request')
-
-        requests = PurchaseRequest.browse(Transaction().context['active_ids'])
-        for request in requests:
+        for request in self.records:
             request.purchase_line = None
-        PurchaseRequest.update_state(requests)
+        self.model.reset_purchased(self.records)
         return 'end'
 
     def transition_cancel_request(self):
-        pool = Pool()
-        PurchaseRequest = pool.get('purchase.request')
-
-        requests = PurchaseRequest.browse(Transaction().context['active_ids'])
-        for request in requests:
+        for request in self.records:
             request.exception_ignored = True
-        PurchaseRequest.update_state(requests)
+        self.model.update_state(self.records)
         return 'end'
 
 
